# -*- coding: utf-8 -*-
from __future__ import absolute_import

import copy
import json
import pysolr
import re
import requests
import time

from celery import shared_task
from lxml import html

from django.conf import settings

from .models import UFGDCombination, ScrapingTime
from search_server.models import Library

UEMS = Library.objects.get(id=1)
UFGD = Library.objects.get(id=2)


def check_and_save(book):
    """
    Check if book is not saved yet, and saves
    :param book: Book parsed in dict format
    :return: True or False, if saves
    """
    if book:
        solr = pysolr.Solr(settings.SOLR_BASEPATH + settings.SOLR_CORE)
        timeout = 1
        while True:
            try:
                solr.add([book])
                return True
            except (requests.exceptions.ConnectionError, pysolr.SolrError) as exception:
                time.sleep(timeout)
                timeout += 0.1
    else:
        return False


def UFGD_save(books):
    """
    Check if book is not saved yet, and saves
    :param book: Book parsed in dict format
    :return: True or False, if saves
    """
    timeout = 1
    solr = pysolr.Solr(settings.SOLR_BASEPATH + settings.SOLR_CORE)
    while True:
        try:
            solr.add(books)
            return True
        except (requests.exceptions.ConnectionError, pysolr.SolrError) as exception:
            time.sleep(timeout)
            timeout += 0.1


def requests_get(url='', headers='', timeout=0.5, retry=True):
    """
    Perform the request with GET method.
    :param url: url to request
    :param headers: headers to request
    :param timeout: timeout to purge request
    :param retry: if True, make retry after time.sleep
    :return: response of page
    """
    exception = True
    response = None
    while exception:
        try:
            response = requests.get(url, headers=headers, timeout=timeout)
            exception = False
        except requests.exceptions.ConnectionError as exception:
            if retry:
                time.sleep(1)
            else:
                exception = False
        except requests.exceptions.Timeout as exception:
            if retry:
                timeout += 0.1
            else:
                exception = False

    return response


@shared_task
def uems_scraping_books(timeout, code):
    """
    Performs requests on the library website via the get method and a
    range of books identifiers, extracts the contents of the responses
    and store in database.
    :param timeout: timeout to purge request
    :param code: number of book in library
    :return: save book in database
    """

    url = 'http://www.uems.br/biblioteca/informacoes.php?Codigo=%s&Local=1'
    path = "//table[2]/tr[%s]/td[%s]//b/text()"
    location_path = "//table[5]/tr"

    time_rq = ScrapingTime()
    time_rq.library = u"UEMS - Dourados"
    time_rq.initial_time = time.time()
    start_time = time.time()
    time_rq.is_valid = False

    response = requests_get(url % str(code), timeout=timeout)

    tree = html.fromstring(response.content)
    # Extract title via xpath
    title = ''.join(tree.xpath(path % (1, 1)))
    location = tree.xpath(location_path)
    # Checks if exists title, if exist gets all book data
    if title != u' ' and location:
        book = copy.deepcopy(settings.BOOK)

        book['title'] = title
        book['author'] = ''.join(tree.xpath(path % (2, 1)))
        local = ''.join(tree.xpath(path % (5, 2)))
        publishing_house = ''.join(tree.xpath(path % (5, 1)))
        year = ''.join(tree.xpath(path % (7, 3)))
        description = ''.join(tree.xpath(path % (7, 1)))
        description = ''.join(re.findall('(\s?[0-9]+p.)', description))
        description = re.sub('\s', '', description)

        if local:
            book['publication'] = local.strip()
            if publishing_house:
                book['publication'] += ': ' + publishing_house
                if year:
                    book['publication'] += ', ' + year
                    if description:
                        book['publication'] += '. ' + description

        edition = ''.join(tree.xpath(path % (7, 2)))
        book['cdd'] = ''.join(tree.xpath(path % (6, 2)))
        book['pha'] = ''.join(tree.xpath(path % (6, 3)))
        book['id_uems'] = code
        book['library'] = UEMS.id


        if book['cdd'].isalpha():
            book['kind'] = u"Monografia/TCC"
            book['area_of_knowledge'] = edition.strip(',')
        else:
            book['kind'] = u"Livro"
            book['edition'] = edition
            book['boost'] = 2.0

        book['id'] = str(hash(frozenset(book.items())))
        clear_book_fields(book)
        if check_and_save(book):
            time_rq.is_valid = True
            time_rq.end_time = time.time()
            time_rq.request_time = time.time() - start_time
            time_rq.save()


@shared_task
def uems_scraping_periodics(timeout, code):
    """
    Function to perform parser of UEMS periodicts
    :param timeout: timeout to purge request
    :param code: number of periodic in the library
    :return: Save book in database
    """
    url = 'http://www.uems.br/biblioteca/informacoes_peri.php?Codigo=%s&Local=1'

    time_rq = ScrapingTime()
    time_rq.library = u"UEMS - Dourados"
    time_rq.initial_time = time.time()
    start_time = time.time()
    time_rq.is_valid = False
    path = "//table[2]/font/tr[%s]/td[%s]//b/text()"

    response = requests_get(url % str(code), timeout=timeout)

    tree = html.fromstring(response.content)
    # Extract title via xpath
    title = ''.join(tree.xpath(path % (1, 1)))

    if title != u' ':
        book = copy.deepcopy(settings.BOOK)

        book['title'] = title
        book['periodicity'] = ''.join(tree.xpath(path % (3, 1)))
        book['country'] = ''.join(tree.xpath(path % (3, 2)))
        book['language'] = ''.join(tree.xpath(path % (3, 3)))
        book['editor'] = ''.join(tree.xpath(path % (4, 1)))
        book['cnn'] = ''.join(tree.xpath(path % (4, 3)))
        book['collections'] = ''.join(tree.xpath('//table[4]/font/tr/td[1]/text()'))
        book['kind'] = u"Periódico"
        book['id_uems'] = code
        book['library'] = UEMS.id
        book['id'] = str(hash(frozenset(book.items())))

        if book['cnn']:
            clear_book_fields(book)
            if check_and_save(book):
                time_rq.is_valid = True
                time_rq.end_time = time.time()
                time_rq.request_time = time.time() - start_time
                time_rq.save()


def clear_book_fields(book):
    """
    Clear fields in book dict, remove unecessary whitespaces
    :param book: dict object
    :return: book dict cleaned
    """
    for key in book:
        if isinstance(book[key], (str, unicode)):
            book[key] = re.sub(r'\s+', ' ', re.sub(r'\s+([,.;:!?"][^$])', r'\1', book[key].strip()))


def ufgd_requests_post(url, payload, headers, timeout):
    """
    Perform the request with POST method.
    :param url: URL to make request POST
    :param payload: params to send in POST request
    :param headers: headers to send with request
    :param timeout: timeout to purge request
    :return: response of page
    """
    exception = True
    search_page = None
    while exception:
        try:
            search_page = requests.post(url, params=payload, headers=headers, timeout=timeout)
            exception = False
        except requests.exceptions.ConnectionError as exception:  # This is the correct syntax
            time.sleep(1)
        except requests.exceptions.Timeout as exception:
            timeout += 0.1

    return search_page


def ufgd_get_cookie(timeout, retry=True):
    """
    Get page cookie from biblioteca.ufgd page
    :param timeout: timeout to purge request
    :param retry: retry request if True
    :return: cookie of library JSESSIONID
    """

    url = 'http://biblioteca.ufgd.edu.br/faces/consulta.xhtml'

    page = requests_get(url, "", timeout, retry)
    aux = str(page.cookies).split(' ')
    cookie = str([s for s in aux if 'JSESSIONID' in s][0])  # get JSESSIONID cookie
    settings.COOKIE = cookie
    return cookie


def parser_ufgd(cont, payload, headers, timeout, url, term, element_type, books):
    """
    Primary function to make requests and parsers of element
    :param cont: number of element in search results page
    :param payload: params to request page
    :param headers: headers to request page
    :param timeout: timeout to purge request
    :param url: URL to perform request
    :param term: term to search
    :param element_type: type of elements in response of search
    :return: function to request elements of result
    """
    time_rq = ScrapingTime()
    time_rq.library = "UFGD - Central"
    time_rq.initial_time = time.time()
    start_time = time.time()
    time_rq.is_valid = True

    payload['javax.faces.source'] = settings.SEARCH_LISTA % str(cont)
    for key, value in payload.items():
        reg = r'^(search:lista:)-?\d+(:%s)$' % (settings.J_IDT)
        if re.search(reg, key):
            payload.pop(key)

    payload[settings.SEARCH_LISTA % str(cont)] = settings.SEARCH_LISTA % str(cont)

    response_modal = ufgd_requests_post(url, payload, headers, timeout)

    modal_tree = html.fromstring(response_modal.content)

    book = copy.deepcopy(settings.BOOK)
    # parse authors
    aux = modal_tree.xpath('//a[@class="linkAutor"]/text()')
    authors = ' '.join(aux)
    authors = re.sub('\n', '', authors.strip())
    book['author'] = authors

    ok = 0
    for i in range(0, 10):
        # parse title
        if '\nTitulo: ' in modal_tree.xpath(('//tr[@id="formMaterial:tableMat_row_%d"]/td[1]/label//text()' % i)):
            aux = modal_tree.xpath(
                ('//tr[@id="formMaterial:tableMat_row_%d"]/td[2]/label//text()' % i)
            )
            title = ' '.join(aux)
            title = re.sub('\n', '', title.strip())
            book['title'] = title
            ok += 1
        # parse location
        elif u'\nLocalização: ' in modal_tree.xpath(
                ('//tr[@id="formMaterial:tableMat_row_%d"]/td[1]/label//text()' % i)):
            aux = modal_tree.xpath(
                ('//tr[@id="formMaterial:tableMat_row_%d"]/td[2]/label//text()' % i)
            )
            location = ' '.join(aux)
            location = re.sub('\n', '', location.strip())
            locations = location.split(' ')
            try:
                book['cdd'] = locations[0]
            except:
                pass
            try:
                book['pha'] = locations[1]
            except:
                pass
            ok += 1
        # parse edition
        elif u'\nEdição: ' in modal_tree.xpath(('//tr[@id="formMaterial:tableMat_row_%d"]/td[1]/label//text()' % i)):
            aux = modal_tree.xpath(
                ('//tr[@id="formMaterial:tableMat_row_%d"]/td[2]/label//text()' % i)
            )
            edition = ' '.join(aux)
            edition = re.sub('\n', '', edition.strip())
            book['edition'] = edition
            ok += 1
        # parse publication
        elif u'\nPublicação: ' in modal_tree.xpath(
                ('//tr[@id="formMaterial:tableMat_row_%d"]/td[1]/label//text()' % i)):
            aux = modal_tree.xpath(
                ('//tr[@id="formMaterial:tableMat_row_%d"]/td[2]/label//text()' % i)
            )
            publication = ' '.join(aux)
            publication = re.sub('\n', '', publication.strip())
            book['publication'] = publication
            ok += 1

        if ok == 4:
            break

    # type_doc
    type_doc = re.sub('[\[\]]', '', element_type)  # remove []
    type_doc = re.sub('\n', '', type_doc.strip())
    type_doc = re.sub('/ ', '/', type_doc)
    if type_doc not in settings.KINDS:
        book['kind'] = u'Outro'
    elif type_doc == "Tese":
        book['kind'] = u'Tese/Dissertação'
    else:
        book['kind'] = type_doc

    book['kind'] = type_doc
    if book['kind'] == u'Livro':
        book['boost'] = 2.0
    book['library'] = UFGD.id

    book['id'] = str(hash(frozenset(book.items())))

    if book['title']:
        clear_book_fields(book)

        combination = UFGDCombination.objects.get(combination=term)
        books.append(book)

        combination.geted_results += 1
        combination.save()

        time_rq.end_time = time.time()
        time_rq.request_time = time.time() - start_time
        time_rq.save()


@shared_task
def ufgd_scraping_books(term, cookie, timeout):
    """
    Using the 'combinations_ufgd.txt' file, performs requests and
    parse the results of the searches with the combinations.
    Collects all registered elements in the UFGD system.
    :param term: term for search in library page
    :param cookie: cookie parsed in first request
    :param timeout: timeout to purge scrapy
    :return: function to parse elements in search result page
    """

    url = 'http://biblioteca.ufgd.edu.br/faces/consulta_resultado.xhtml?filtro=conteudo&termo=%s' % (term)

    # headers to requests
    headers = copy.deepcopy(settings.HEADERS)
    headers['Cookie'] = cookie
    # params to request results of search
    payload = copy.deepcopy(settings.PAYLOAD)
    # params to request page of search
    payload_page = copy.deepcopy(settings.PAYLOAD_PAGE)

    # set headers to request search
    headers.pop('Faces-Request')
    headers['Referer'] = url

    page = requests_get(url, headers, timeout)

    # transforming response in tree
    tree = html.fromstring(page.content)

    # parsing javax.faces.ViewState code
    cod = tree.xpath('//*[@id="javax.faces.ViewState"]/@value')
    cod = str(cod[0])
    payload_page['javax.faces.ViewState'] = cod
    payload['javax.faces.ViewState'] = cod

    # parsing results quantity
    qnt = tree.xpath('//label[@id="search:qtdade"]//text()')
    try:
        qnt = map(int, re.findall('\d+', qnt[0]))[0]  # convert to number
    except:
        qnt = 0

    payload_page['search:lista_rows'] = str(qnt)

    headers['Faces-Request'] = 'partial/ajax'

    # setting term to search
    payload_page['search:termo2'] = term
    payload['search:termo2'] = term

    if qnt > 0:
        # get search page to do pagination
        pagination_page = ufgd_requests_post(url, payload_page, headers, timeout)

        # parsing genero
        elements_type = html.fromstring(pagination_page.content).xpath(
            '//label[@class="destaque-genero"]//text()'
        )

        # loop into results of search
        books = []
        for cont in range(0, qnt):
            parser_ufgd(cont, payload, headers, timeout, url, term, elements_type[cont], books)
        UFGD_save(books)


@shared_task
def ufgd_scraping_comb(timeout, comb):
    """
    Test all possible combinations, check if exists results with that.
    Save the combination, number of results and repeated number for
    each combination.
    :param timeout: timeout to purge request comb
    :param comb: combination to try request
    :return: function to parse result page
    """
    time.sleep(1)
    url = 'http://biblioteca.ufgd.edu.br/faces/consulta_resultado.xhtml?filtro=conteudo&termo=%s'

    cookie = settings.COOKIE if settings.COOKIE else ufgd_get_cookie(timeout, retry=True)

    try:  # check if comb arealy exists
        db_comb = UFGDCombination.objects.get(combination=comb)
    except:
        db_comb = None

    # access text in label with xpath
    page = requests_get(url % comb, "", timeout)
    tree = html.fromstring(page.content)

    # access text in label with xpath
    results_number = tree.xpath('//label[@id="search:qtdade"]//text()')

    if results_number:
        number = map(int, re.findall('\d+', results_number[0]))[0]  # convert to number
        if number < 2000:
            if not db_comb:
                combinations = UFGDCombination()
                combinations.combination = comb
                combinations.results_number = number
                combinations.repeated_results = 0
                combinations.geted_results = 0
                combinations.save()
        else:
            if db_comb:
                db_comb.delete()
            ufgd_insert_letters(timeout, comb)
    elif tree.xpath('//*[@id="search:mensagens"]/div/ul/li/span'):
        try:
            db_comb.delete()
        except:
            pass
        ufgd_insert_letters(timeout, comb)


@shared_task
def ufgd_insert_letters(timeout, comb):
    """
    Insert new list of letters to specific combination.
    :param timeout: timeout to purge request
    :param comb: combination to add in queue of combinations to search
    :return: list with more combinations appended
    """

    for letter in range(97, 123):
        combination = comb[:len(comb) - 1] + str(unichr(letter)) + comb[len(comb) - 1:]
        ufgd_scraping_comb.apply_async(args=[timeout, combination])


@shared_task
def ufgd_scraping(timeout):
    """
    Generate the first list of combinations with all alphabet letters and *.
    For each initial combination, test if return more than 2000 results,
    if yes, generate new list of sub combinations to that initial combination.
    :param timeout: timeout to purge requests
    :return: asynchronous task to add combination, perform requests and parse
    """

    for combination in range(97, 123):
        comb = ('%s*' % str(unichr(combination)))
        ufgd_scraping_comb.apply_async(args=[timeout, comb])


def book_avaliable_uems(timeout, cod):
    """
    Request to parse total and avaliable unitys element in UEMS library
    :param timeout: timeout to purge request
    :param cod: code to search
    :return: json object like {'avaliables': X, 'total': X}
    """
    url = 'http://www.uems.br/biblioteca/informacoes.php?Codigo=%d&Local=1'
    page = requests_get(url % int(cod), "", timeout, retry=False)
    tree = html.fromstring(page.content)
    result = {}
    try:
        result['total'] = int(''.join(re.findall(r'(\d+)', ''.join(tree.xpath('//table[5]/tr[1]/td[1]/text()')).strip())))
        result['avaliables'] = int(''.join(re.findall(r'(\d+)', ''.join(tree.xpath('//table[5]/tr[1]/td[4]/text()')).strip())))
    except:
        pass
    return json.dumps(result)


def book_avaliable_ufgd(timeout, term):
    """
    Request to parse total and avaliable unitys element in UFGD library
    :param timeout: timeout to purge request
    :param term: term to search element like "CDD PHA"
    :return: json object like {'avaliables': X, 'total': X}
    """
    url = 'http://biblioteca.ufgd.edu.br/faces/consulta_resultado.xhtml?filtro=conteudo&termo=%s'
    cookie = ufgd_get_cookie(timeout, retry=False)
    page = requests_get(url=(url % term), timeout=timeout, retry=False)
    tree = html.fromstring(page.content)
    results_number = tree.xpath('//label[@id="search:qtdade"]//text()')

    if not results_number or map(int, re.findall(r'\d+', results_number[0]))[0] > 1:
        return {}

    # headers to requests
    headers = copy.deepcopy(settings.HEADERS)
    headers['Cookie'] = cookie
    # params to request results of search
    payload = copy.deepcopy(settings.PAYLOAD)
    url = url % term

    # set headers to request search
    headers.pop('Faces-Request')
    headers['Referer'] = url

    page = requests_get(url, headers, timeout, retry=False)

    # transforming response in tree
    tree = html.fromstring(page.content)

    # parsing javax.faces.ViewState code
    cod = tree.xpath('//*[@id="javax.faces.ViewState"]/@value')
    cod = str(cod[0])
    payload['javax.faces.ViewState'] = cod

    headers['Faces-Request'] = 'partial/ajax'

    # setting term to search
    payload['search:termo2'] = term

    for key, value in payload.items():
        reg = r'^(search:lista:)-?\d+(:%s)$' % (settings.J_IDT)
        if re.search(reg, key):
            payload.pop(key)
    payload[settings.SEARCH_LISTA % str(0)] = settings.SEARCH_LISTA % str(0)

    response_modal = ufgd_requests_post(url, payload, headers, timeout)

    modal_tree = html.fromstring(response_modal.content)
    result = {}
    try:
        result['avaliables'] = int(
            modal_tree.xpath('count(//tbody[@id="formMaterial:tableEx_data"]/tr/td/img[@src="/resources/img/disponivel_peq.png"])'))
        result['total'] = int(
            ''.join(re.findall(r'\d+$', ''.join(modal_tree.xpath('//*[@id="formMaterial:tableEx"]/div/div/label/text()')))))
    except:
        pass

    return json.dumps(result)
